import ast
from typing import Iterator, Optional

__all__ = ["relate", "get_parents", "first_occurrence"]


def relate(tree: ast.AST, parent: Optional[ast.AST] = None) -> None:
    tree.parent = parent  # type: ignore
    for node in ast.walk(tree):
        for child in ast.iter_child_nodes(node):
            child.parent = node  # type: ignore


def get_parents(node: ast.AST) -> Iterator[ast.AST]:
    parent = node
    while parent:
        parent = parent.parent  # type: ignore
        if parent:
            yield parent


def first_occurrence(node: ast.AST, *ancestors):
    for parent in get_parents(node):
        if isinstance(parent, ancestors):
            return parent
    else:
        return None



from typing import List, Optional, Sequence, TypeVar, Union, cast

import libcst as cst
import libcst.matchers as m
from libcst.metadata import CodeRange, PositionProvider

from unimport.statement import Import, ImportFrom

__all__ = ["refactor_string"]

ImportT = TypeVar("ImportT", bound=Union[cst.Import, cst.ImportFrom])


class _RemoveUnusedImportTransformer(cst.CSTTransformer):
    METADATA_DEPENDENCIES = (PositionProvider,)

    def __init__(
        self, unused_imports: List[Union[Import, ImportFrom]]
    ) -> None:
        self.unused_imports = unused_imports

    @staticmethod
    def get_import_name_from_attr(attr_node: cst.Attribute) -> str:
        name = [attr_node.attr.value]  # last value
        node = attr_node.value
        while m.matches(node, m.OneOf(m.Name(), m.Attribute())):
            if isinstance(node, cst.Attribute):
                name.append(node.attr.value)
                node = node.value
            else:
                name.append(cst.ensure_type(node, cst.Name).value)
                break
        name.reverse()
        return ".".join(name)

    def is_import_used(
        self, import_name: str, column: int, location: CodeRange
    ) -> bool:
        for imp in self.unused_imports:
            if all(
                (
                    imp.name == import_name,
                    imp.lineno == location.start.line,
                    imp.column == column,
                )
            ):
                return False
        return True

    def get_location(
        self, node: Union[cst.Import, cst.ImportFrom]
    ) -> CodeRange:
        return self.get_metadata(cst.metadata.PositionProvider, node)

    @staticmethod
    def get_rpar(
        rpar: Optional[cst.RightParen], location: CodeRange
    ) -> Optional[cst.RightParen]:
        if not rpar or location.start.line == location.end.line:
            return rpar
        else:
            return cst.RightParen(
                whitespace_before=cst.ParenthesizedWhitespace()
            )

    def leave_import_alike(
        self,
        original_node: ImportT,
        updated_node: ImportT,
    ) -> Union[cst.RemovalSentinel, ImportT]:
        names_to_keep = []
        names = cast(Sequence[cst.ImportAlias], updated_node.names)
        # already handled by leave_ImportFrom
        for column, import_alias in enumerate(names):
            if isinstance(import_alias.name, cst.Attribute):
                import_name = self.get_import_name_from_attr(
                    attr_node=import_alias.name
                )
            else:
                raw_import = import_alias.asname or import_alias
                raw_import_name = cst.ensure_type(raw_import.name, cst.Name)
                import_name = raw_import_name.value
            if self.is_import_used(
                import_name, column + 1, self.get_location(original_node)
            ):
                names_to_keep.append(import_alias)
        if not names_to_keep:
            return cst.RemoveFromParent()
        elif len(names) == len(names_to_keep):
            return updated_node
        else:
            names_to_keep[-1] = names_to_keep[-1].with_changes(
                comma=cst.MaybeSentinel.DEFAULT
            )
            return cast(
                ImportT, updated_node.with_changes(names=names_to_keep)
            )

    @staticmethod
    def leave_StarImport(
        updated_node: cst.ImportFrom,
        imp: ImportFrom,
    ) -> Union[cst.ImportFrom, cst.RemovalSentinel]:
        if imp.suggestions:
            names_to_suggestions = [
                cst.ImportAlias(cst.Name(module)) for module in imp.suggestions
            ]
            return updated_node.with_changes(names=names_to_suggestions)
        else:
            return cst.RemoveFromParent()

    def leave_Import(
        self, original_node: cst.Import, updated_node: cst.Import
    ) -> Union[cst.RemovalSentinel, cst.Import]:
        return self.leave_import_alike(original_node, updated_node)

    def leave_ImportFrom(
        self, original_node: cst.ImportFrom, updated_node: cst.ImportFrom
    ) -> Union[cst.RemovalSentinel, cst.ImportFrom]:
        if isinstance(updated_node.names, cst.ImportStar):

            def get_star_imp() -> Optional[ImportFrom]:
                if isinstance(updated_node.module, cst.Attribute):
                    import_name = self.get_import_name_from_attr(
                        attr_node=updated_node.module
                    )
                else:
                    import_name = updated_node.module.value
                location = self.get_location(original_node)
                for imp in self.unused_imports:
                    if (
                        isinstance(imp, ImportFrom)
                        and imp.name == import_name
                        and imp.lineno == location.start.line
                    ):
                        return imp
                else:
                    return None

            imp = get_star_imp()
            if imp:
                return self.leave_StarImport(updated_node, imp)
            else:
                return original_node
        rpar = self.get_rpar(
            updated_node.rpar, self.get_location(original_node)
        )
        return self.leave_import_alike(
            original_node, updated_node.with_changes(rpar=rpar)
        )


def refactor_string(
    source: str,
    unused_imports: List[Union[Import, ImportFrom]],
) -> str:
    if unused_imports:
        wrapper = cst.MetadataWrapper(cst.parse_module(source))
        fixed_module = wrapper.visit(
            _RemoveUnusedImportTransformer(unused_imports)
        )
        return fixed_module.code
    return source



import ast
import builtins
import distutils.sysconfig
import sys
from typing import Any, Callable, TypeVar, Union

from unimport.statement import Import, ImportFrom

__all__ = [
    "DESCRIPTION",
    "VERSION",
    "PY38_PLUS",
    "SUBSCRIPT_TYPE_VARIABLE",
    "IGNORE_IMPORT_NAMES",
    "BUILTINS",
    "Function",
    "ASTImportableT",
    "ASTFunctionT",
    "ImportT",
    "CFNT",
    "DefTuple",
    "ASTFunctionTuple",
    "BUILTIN_MODULE_NAMES",
    "STDLIB_PATH",
    "GLOB_PATTERN",
    "INCLUDE_REGEX_PATTERN",
    "EXCLUDE_REGEX_PATTERN",
]

DESCRIPTION = (
    "A linter, formatter for finding and removing unused import statements."
)
VERSION = "0.7.1"

PY38_PLUS = sys.version_info >= (3, 8)

SUBSCRIPT_TYPE_VARIABLE = frozenset(
    {
        "AbstractSet",
        "AsyncContextManager",
        "AsyncGenerator",
        "AsyncIterable",
        "AsyncIterator",
        "Awaitable",
        "Callable",
        "ChainMap",
        "ClassVar",
        "Collection",
        "Container",
        "ContextManager",
        "Coroutine",
        "Counter",
        "DefaultDict",
        "Deque",
        "Dict",
        "FrozenSet",
        "Generator",
        "IO",
        "ItemsView",
        "Iterable",
        "Iterator",
        "KeysView",
        "List",
        "Mapping",
        "MappingView",
        "Match",
        "MutableMapping",
        "MutableSequence",
        "MutableSet",
        "Optional",
        "Pattern",
        "Reversible",
        "Sequence",
        "Set",
        "SupportsRound",
        "Tuple",
        "Type",
        "Union",
        "ValuesView",
        # Python >= 3.7.
        "Literal",
        # Python >= 3.8.
        "OrderedDict",
    }
)

IGNORE_IMPORT_NAMES = frozenset({"__all__", "__doc__", "__name__"})
BUILTINS = frozenset(dir(builtins))

Function = TypeVar("Function", bound=Callable[..., Any])
ASTImportableT = Union[
    ast.AsyncFunctionDef,
    ast.Attribute,
    ast.ClassDef,
    ast.FunctionDef,
    ast.Name,
    ast.alias,
]

ASTFunctionT = Union[ast.FunctionDef, ast.AsyncFunctionDef]
ImportT = Union[Import, ImportFrom]
CFNT = Union[ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef, ast.Name]
DefTuple = (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)
ASTFunctionTuple = (ast.FunctionDef, ast.AsyncFunctionDef)
BUILTIN_MODULE_NAMES = frozenset(sys.builtin_module_names)
STDLIB_PATH = distutils.sysconfig.get_python_lib(standard_lib=True)
GLOB_PATTERN = "**/*.py"
INCLUDE_REGEX_PATTERN = "\\.(py)$"
EXCLUDE_REGEX_PATTERN = "^$"



"""It offers some utils."""
import contextlib
import difflib
import functools
import importlib.machinery
import importlib.util
import re
import tokenize
from distutils.util import strtobool
from pathlib import Path
from typing import FrozenSet, Iterable, Iterator, List, Optional, Set, Tuple

from importlib_metadata import PackageNotFoundError, metadata
from pathspec.patterns.gitwildmatch import GitWildMatchPattern

from unimport import constants as C

__all__ = [
    "get_dir",
    "get_source",
    "get_spec",
    "is_std",
    "package_name_from_metadata",
    "get_used_packages",
    "actiontobool",
    "get_exclude_list_from_gitignore",
    "read",
    "list_paths",
    "diff",
]


def get_dir(package: str) -> FrozenSet[str]:
    try:
        module = importlib.import_module(package)
    except (ImportError, AttributeError, TypeError, ValueError):
        return frozenset()
    return frozenset(dir(module))


def get_source(package: str) -> Optional[str]:
    spec = get_spec(package)
    # The below two can be one of several values per their previous type annotations
    # But for our use case, we know that these are the specific classes that asserted to be
    if spec:
        assert isinstance(spec.loader, importlib.machinery.SourceFileLoader)
        assert isinstance(spec.loader.path, str)
        if spec.loader.path.endswith(".py"):
            return spec.loader.get_data(spec.loader.path).decode("utf-8")
    return None


@functools.lru_cache(maxsize=None)
def get_spec(package: str) -> Optional[importlib.machinery.ModuleSpec]:
    try:
        return importlib.util.find_spec(package)
    except (ImportError, AttributeError, TypeError, ValueError):
        return None


def is_std(package: str) -> bool:
    """Returns True if package module came with from Python."""

    if package in C.BUILTIN_MODULE_NAMES:
        return True
    spec = get_spec(package)
    if spec and isinstance(spec.origin, str):
        return any(
            (
                spec.origin.startswith(C.STDLIB_PATH),
                spec.origin in ["built-in", "frozen"],
                spec.origin.endswith(".so"),
            )
        )
    else:
        return False


@functools.lru_cache(maxsize=None)
def package_name_from_metadata(package: str) -> Optional[str]:
    if not is_std(package):
        with contextlib.suppress(PackageNotFoundError):
            return metadata(package)["Name"]
    return None


def get_used_packages(
    imports: List[C.ImportT], unused_imports: List[C.ImportT]
) -> Set[str]:
    packages = set()
    used_packages = set(
        map(lambda imp: imp.package.split(".")[0], imports)
    ) - set(map(lambda imp: imp.package.split(".")[0], unused_imports))
    for package in used_packages:
        name = package_name_from_metadata(package)
        if name:
            packages.add(name)
    return packages


def actiontobool(action: str) -> bool:
    if action == "":
        return True
    with contextlib.suppress(ValueError):
        return strtobool(action)
    return False


def get_exclude_list_from_gitignore() -> List[str]:
    """Converts .gitignore patterns to regex and return this exclude regex
    list."""
    path = Path(".gitignore")
    gitignore_regex: List[str] = []
    if path.is_file():
        source, _ = read(path)
        for line in source.splitlines():
            regex = GitWildMatchPattern.pattern_to_regex(line)[0]
            if regex:
                gitignore_regex.append(regex)
    return gitignore_regex


def read(path: Path) -> Tuple[str, str]:
    try:
        with tokenize.open(path) as stream:
            source = stream.read()
            encoding = stream.encoding
    except (OSError, SyntaxError) as err:
        return "", "utf-8"
    return source, encoding


def list_paths(
    start: Path,
    include: str = C.INCLUDE_REGEX_PATTERN,
    exclude: str = C.EXCLUDE_REGEX_PATTERN,
) -> Iterator[Path]:
    include_regex, exclude_regex = re.compile(include), re.compile(exclude)
    file_names: Iterable[Path]
    if start.is_dir():
        file_names = start.glob(C.GLOB_PATTERN)
    else:
        file_names = [start]
    yield from filter(
        lambda filename: include_regex.search(str(filename))
        and not exclude_regex.search(str(filename)),
        file_names,
    )


def diff(
    *, source: str, refactor_result: str, fromfile: Path = None
) -> Tuple[str, ...]:
    return tuple(
        difflib.unified_diff(
            source.splitlines(),
            refactor_result.splitlines(),
            fromfile=fromfile.as_posix() if fromfile else "",
        )
    )



import operator
from typing import List, NamedTuple, Union

__all__ = ["Import", "ImportFrom", "Name"]


class Import(NamedTuple):
    lineno: int
    column: int
    name: str
    package: str

    def __len__(self) -> int:
        return operator.length_hint(self.name.split("."))

    def is_match_sub_packages(self, name_name) -> bool:
        return self.name.split(".")[0] == name_name


class ImportFrom(NamedTuple):
    lineno: int
    column: int
    name: str
    package: str
    star: bool
    suggestions: List[str]

    def __len__(self) -> int:
        return operator.length_hint(self.name.split("."))

    def is_match_sub_packages(self, name_name):
        return False


class Name(NamedTuple):
    lineno: int
    name: str
    is_all: bool = False

    @property
    def is_attribute(self):
        return "." in self.name

    def match(self, imp: Union[Import, ImportFrom]) -> bool:
        if self.is_all:
            return self.name == imp.name
        elif self.is_attribute:
            return imp.lineno < self.lineno and (
                ".".join(self.name.split(".")[: len(imp)]) == imp.name
                or imp.is_match_sub_packages(self.name)
            )
        else:
            return (imp.lineno < self.lineno) and (
                self.name == imp.name or imp.is_match_sub_packages(self.name)
            )



import configparser
from ast import literal_eval
from pathlib import Path
from typing import List, NamedTuple

import toml

from unimport import constants as C
from unimport import utils

__all__ = ["DefaultConfig", "Config"]

CONFIG_FILES = {"setup.cfg": "unimport", "pyproject.toml": "tool.unimport"}


class DefaultConfig(NamedTuple):
    sources: List[Path] = [Path(".")]
    include: str = C.INCLUDE_REGEX_PATTERN
    exclude: str = C.EXCLUDE_REGEX_PATTERN
    requirements: bool = False
    gitignore: bool = False
    remove: bool = False
    diff: bool = False
    include_star_import: bool = False
    permission: bool = False
    check: bool = False

    def merge(self, **kwargs):
        diff_dict = set(kwargs) - set(self._asdict())
        # delete keys that are not available.
        for invalid_key in diff_dict:
            del kwargs[invalid_key]
        # delete items if they are the same as default values
        for key, value in kwargs.copy().items():
            if getattr(self, key) == value:
                del kwargs[key]
        config = self._replace(**kwargs)
        diff = kwargs.get("diff") or kwargs.get("permission")
        config = config._replace(
            diff=diff or any((config.diff, config.permission))
        )
        config = config._replace(
            check=kwargs.get("check") or not any((config.diff, config.remove))
        )
        if config.gitignore:
            gitignore_exclude = utils.get_exclude_list_from_gitignore()
            gitignore_exclude.extend(config.exclude)
            config = config._replace(exclude="|".join(gitignore_exclude))
        return config


class Config:

    default_config = DefaultConfig()

    def __init__(self, config_file: Path) -> None:
        self.config_file = config_file
        self.section = CONFIG_FILES[config_file.name]

    def parse(self) -> DefaultConfig:
        return getattr(self, f"parse_{self.config_file.suffix.strip('.')}")()

    def parse_cfg(self) -> DefaultConfig:
        parser = configparser.ConfigParser(allow_no_value=True)
        parser.read(self.config_file)
        if parser.has_section(self.section):

            def get_config_as_list(name: str) -> List[str]:
                return literal_eval(
                    parser.get(
                        self.section,
                        name,
                        fallback=getattr(self.default_config, name),
                    )
                )

            cfg_context = {}
            config_annotations = self.default_config.__annotations__
            for key, value in parser[self.section].items():
                key_type = config_annotations[key]
                if key_type == bool:
                    cfg_context[key] = parser.getboolean(self.section, key)
                elif key_type == str:
                    cfg_context[key] = value  # type: ignore
                elif key_type == List[Path]:
                    cfg_context[key] = [  # type: ignore
                        Path(p) for p in get_config_as_list(key)
                    ]
            return self.default_config._replace(**cfg_context)  # type: ignore
        else:
            return self.default_config

    def parse_toml(self) -> DefaultConfig:
        parsed_toml = toml.loads(self.config_file.read_text())
        config = parsed_toml.get("tool", {}).get("unimport", {})
        sources = config.get("sources", self.default_config.sources)
        config["sources"] = [Path(path) for path in sources]
        return self.default_config._replace(**config)



import sys
from pathlib import Path

from setuptools import setup

import unimport.constants as C

assert sys.version_info >= (3, 6, 0), "unimport requires Python 3.6+"

CURRENT_DIR = Path(__file__).parent


def get_long_description():
    readme_md = CURRENT_DIR / "README.md"
    with open(readme_md, encoding="utf8") as ld_file:
        return ld_file.read()


setup(
    name="unimport",
    version=C.VERSION,
    description=C.DESCRIPTION,
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    keywords=["unused", "import"],
    author="Hakan Çelik",
    author_email="hakancelik96@outlook.com",
    url="https://github.com/hakancelik96/unimport",
    project_urls={
        "Documentation": "https://unimport.hakancelik.dev/",
        "Issues": "https://github.com/hakancelik96/unimport/issues",
    },
    license="MIT",
    license_file="LICENSE",
    python_requires=">=3.6.0",
    packages=["unimport"],
    install_requires=[
        "libcst==0.3.13",
        "pathspec==0.8.0",
        "toml==0.10.1",
        "importlib_metadata==2.0.0",
    ],
    extras_require={},
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    entry_points={"console_scripts": ["unimport = unimport.__main__:main"]},
)
